/*
 * Name:    help.c
 * Purpose: display help about commands & system features
 * Syntax:  help [cmd]
 * Comment: Help displays help screens on all user commands
 *          as set up in the file /usr/lib/helpfile.
 *          It is particularly useful on systems with restricted
 *          diskspace, where it is not practical to keep the manuals
 *          on-line.
 *          The command is patterned after the MKS Toolkit command
 *          of the same name. If no /usr/lib/helpindex file exists, or if
 *          it is older than the helpfile, a new helpindex will be
 *          created to allow faster access to the helpscreens by
 *          means of fseek().
 *          Desirable enhancements would include the possibility to
 *          have help look in an environment variable HELPFILE for
 *          an alternate or additional helpfile,
 *              [ Done by John Plocher - uses HELPDIR to set the directory
 *              where the help file and its index are.  It also uses HELPFILE
 *              to give the name of the data file ($HELPFILE.idx is the index)
 *              defaults are "/usr/lib" and "helpfile" ]
 *          and/or to have help look in $HOME/.helpfile if a command
 *          is not found in the default helpfile.
 *          Another useful feature would be the possibility of using
 *          a pager specified by $PAGER to output the help info, since
 *          a few entries (stty, vi, ...) are longer than one screen.
 *          No doubt the coding could be considerably improved, and
 *          any suggestions would be more than welcome.
 * Author:  Wolf Paul, ihnp4!killer!dcs!wnp
 * Date:    Dec. 18, 1987
 *
 *      Revision Control Information
 *
 *      Last edited by: $Author: plocher $
 *                      $Revision: 1.1 $
 *      Last modified:  $Date: 87/12/26 23:34:50 $
 *      Source is in:   $Source: /u/microport/src/help/RCS/help.c,v $
 *
 *      Release state:  $State: Posted $
 *
 * Modification Log
 * ----------------
 *
 *      $Log:   help.c,v $
 *      Revision 1.1  87/12/26  23:34:50  plocher
 *      Added enviroment variables for HELPDIR and HELPFILE
 *
 *  MSDOS - Turbo C modifications    3/28/89  R. Brittain
 *      Changed open mode on index file to "rb" or "wb" instead of "r" or "w"
 *      Changed default directory to /lib
 *      Added code to make index search non case-senstive - lcase and ucase functions
 *      Added  -i option to display index
 *      Added  -k option - code to perform apropos search, added synopsis
 *                to index structure
 *      Added code to specify multiple help files in the HELPFILE variable
 *			   All errors are now non-fatal until all helpfiles in list have
 *			   been processed.  exit() removed from buildindex()
 *      Revised syntax:
 *           set HELPFILE=helpfile1;helpfile2...
 *           help [-i | -k keyword | command]
 *
 *      Compile Turbo C tiny model, no floating point routines, link and convert to .COM
 *
 */

#ifndef lint
    static char rcsid[] =
	 "$Header: help.c,v 2.0 89/04/16 14:30:00 richard Posted $";
#endif

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#define MAXLINE         134
#define SYNOPSIS		80
#define HELPDIR         "/lib"                  /* Added by John Plocher */
#define HELPFILE        "helpfile"              /* .. */
#define INDEX           ".idx"                  /* .. */

char helpfilename[128];        					/* John Plocher */
char helpidxname[128];        				    /* .. */
char helpfilelist[128];
char helpdirname[128];

struct
{
   char name[15];
   char synopsis[SYNOPSIS];
   long offset;
} entry;							   /* helpindex entries for each command */

main(argc, argv)
int argc;
char **argv;
{
	struct stat sbuf1, sbuf2;          /* stat buffers for helpfile & helpindex */
	char 	 *command, *line, Line[MAXLINE], *ucase();
	register char   *cp;               /* John Plocher */
	extern   char   *getenv();         /* .. */
	int		 status, found;
	void	 display_index(), command_apropos();
	FILE 	 *ifp, *hfp;               /* file pointers for helpfile and helpindex */

	if ( argc == 1 )                   /* If no arguments, ... */
		command = "HELP";              /* ... default to "help help" */
	else
		command = ucase(argv[1]);      /* else look for command in argv[1] */

	/*              Added by John Plocher Sat Dec 26 22:02:09 CST 1987
	 * Look for:
	 *      getenv("HELPDIR") + getenv("HELPFILE")
	 *      getenv("HELPDIR") + "helpfile"
	 *      "/usr/lib"                + getenv("HELPFILE")
	 *      "/usr/lib"                + "helpfile"
	 */
	if ((cp = getenv("HELPDIR")) != NULL)   /* if a dir is given, use it */
		strcpy(helpdirname, cp);
	else
		strcpy(helpdirname, HELPDIR);

	strcat(helpdirname, "/");               /* change to "\" if needed on some versions of MSDOS */

	if ((cp = getenv("HELPFILE")) != NULL)  /* if a filename is given, use it */
		strcat(helpfilelist, cp);
	else
		strcat(helpfilelist, HELPFILE);

	found = 0;								/* flag to keep track of success */
	cp = strtok( helpfilelist,";");         /* get first file from list */
	while (cp != (char *) NULL) {

		strcpy(helpfilename, helpdirname);      /* construct full filename */
		strcat(helpfilename, cp);

		strcpy(helpidxname, helpfilename);      /* make a name for the index file */
		strcat(helpidxname, INDEX);

		stat(helpfilename, &sbuf1);       		/* get mtime for helpfile */
		status=access(helpidxname, 0);
		if ( status == 0 )					 	/* if helpindex exists ... */
		{
			stat(helpidxname, &sbuf2);      	/* get mtime for helpindex */
		}
		if ( (status != 0) ||                   /* if there is no helpindex ... */
			((time_t)sbuf1.st_mtime > (time_t)sbuf2.st_mtime) )
												/* or if it is older than helpfile */
		{
			status = buildindex();				/* build a new helpindex */
			if (status) goto get_next_filename;
		}

		if ( (ifp=fopen(helpidxname, "rb")) == NULL )
		{
			fprintf(stderr, "Can't read %s\n", helpidxname);
			goto get_next_filename;
		}
		if (strcmp(command,"-I")==0) {
			if (!found)
				printf("Help is available for the following topics: \n\n");
			display_index(ifp);
			fclose(ifp);
			found++;
			goto get_next_filename; }
		else if (strcmp(command,"-K")==0) {
			if (argc <= 2) {
				fprintf(stderr,"apropos: HELP -k keyword");
				close(ifp);
				exit(1) ; }
			else {
				command_apropos(ifp,ucase(argv[2]));
				fclose(ifp);
				found++;
				goto get_next_filename; }
		}

		while ( 1 )                    /* look for index entry for "command" */
		{
			status=fread(&entry, sizeof(entry), 1, ifp);
			if ( status==0 )           /* quit at end of index file */
			{
				goto get_next_filename;
			}
			if ( strcmp(entry.name, command) == 0 ) /* quit when we find it */
			{
				fclose(ifp);
				found++;
				break;
			}
		}

		if ((hfp=fopen(helpfilename, "r")) == NULL )
		{
			fprintf(stderr, "Can't open %s\n", helpfilename);
			goto get_next_filename;
		}

		fseek(hfp, entry.offset, 0);    /* go to the help entry */

		while ( 1 )                     /* just copy lines to stdout */
		{
			line = fgets(Line, MAXLINE, hfp);
			if ( line == (char *) NULL || line[0] == '#' )
										/* until another entry starts */
				break;
			fputs(line,stdout);
		}
		get_next_filename:;
		cp = strtok(NULL,";");          /* look for another filename in list */
	}
	if (!found) {
		fprintf(stderr, "No help for %s\n", command);
		exit(1);
	}
	fclose(hfp);
	exit(0);
}

int buildindex()
{
	FILE 	*hfp, *ifp;
	char 	Line[MAXLINE], Line2[SYNOPSIS];
	char 	*line, *s, *ucase();
	int 	i,j;

	unlink(helpidxname); /* remove old index file */
	if ( (hfp=fopen(helpfilename, "r")) == NULL )
	{
		fprintf(stderr,"buildindex: Can't read %s\n", helpfilename);
		return (-1);
	}
	if ( (ifp=fopen(helpidxname, "wb")) == NULL )
	{
		fprintf(stderr, "buildindex: Can't write %s\n", helpidxname);
		return(-1);
	}

	while (1)       /* Read thru helpfile ... */
	{
		entry.offset=(long) 0;
		strcpy(entry.synopsis," ");
		line = fgets(Line, MAXLINE, hfp);
		if ( line == (char *) NULL ) break;
		if ( line[0] == '#' )          /* and for each help entry ... */
		{
			entry.offset=ftell(hfp);
			s = fgets(Line2, SYNOPSIS-1, hfp);
			strcpy(entry.synopsis,s);
			line++;
			while ( isspace(line[0]) ) line++;
			i=j=0;
			while ( line[i] != '\0' )
			{
				if ( line[i] == '\n' ) break;
				while ( line[i] == ' ' || line[i] == ',' ) i++;
				while ( !isspace(line[i] ) &&
						line[i] != ',')       /* save its name ... */
				{
					entry.name[j] = toupper(line[i]);
                                        i++; j++;
				}
				while ( j < 15 )
					entry.name[j++] = '\0';
				j = 0;
				fwrite(&entry, sizeof(entry), 1, ifp); /* and write it to indexfile */
			}
		}
	}
	fclose(hfp);
	fclose(ifp);
	return(0);
}

void display_index(ifp)
FILE *ifp;
{
/*  display the index entries, unsorted, 4 to a line */

	int status, i;
	char *lcase();

	while ( 1 ) 					   /* look for index entries */
	{
		for (i=1;i<6;i++) {
			status=fread(&entry, sizeof(entry), 1, ifp);
			if ( status==0 ) {
				printf("\n");
				return;
			}                          /* quit at end of index file */
			printf("%-15s",lcase(entry.name));
		}
		printf("\n");
	}
	return;
}

void command_apropos(ifp,key)
FILE *ifp;
char *key;
{
/*  search the index for synopses matching keyword */

	int  status, i;
	long lastoffset;
	char *lcase();
	char string[SYNOPSIS];

	lastoffset = (long)0;
	while ( 1 )             		   /* look for index entries */
	{
		status=fread(&entry, sizeof(entry), 1, ifp);
		if ( status==0 ) return;       /* quit at end of index file */
		strcpy(string,entry.synopsis);
		if (    (strstr(ucase(string),key) != (char *)NULL)
			 && (lastoffset != entry.offset) )
		{
			printf("%s",entry.synopsis);
			lastoffset = entry.offset;
		}
	}
	return;
}

char *ucase(s)
char *s;
{
	char *c;
	for (c=s;*c!='\0';c++) *c=toupper(*c);
	return(s);
}

char *lcase(s)
char *s;
{
	char *c;
	for (c=s;*c!='\0';c++) *c=tolower(*c);
	return(s);
}
